/*
** Module   : AbstractExtentParameter.java
** Abstract : Base class for special wrapper classes used for variables passed as extent 
**            parameters to functions or procedures.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20140613 Created initial version.
*/

package com.goldencode.p2j.util;

import java.util.*;

import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.BlockManager.Condition;
import com.goldencode.p2j.util.BlockManager.ReturnType;

/**
 * Abstract class for special wrappers used for variables passed as extent parameters 
 * to functions or procedures. The class introduces common logic for all extent parameter cases.
 */
public abstract class AbstractExtentParameter<T extends BaseDataType>
{
   /** Object which manages all instances of this class by context */
   private static final ContextLocal<WorkArea> context = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue() 
      { 
         WorkArea wa = new WorkArea();
         TransactionManager.registerScopeable(wa);
         return wa;
      }
   };

   /** The original variable set to the parameter. */
   private T[] variable = null;
   
   /** The latest parameter reference. */
   private T[] parameter = null;
   
   /**
    * The constructor responsible for instance initialization.
    * <p>
    * The c'tor caches the variable reference in this instance. If {@code null} is 
    * passed to {@code variable}, the value reference is retrieved by calling 
    * {@link #getVariable()}, otherwise the passed-in value is used.
    * 
    * @param   variable
    *          The value reference passed to the function/procedure's parameter
    *          or {@code null}.
    */
   protected AbstractExtentParameter(T[] variable)
   {
      // save the original array, in case it gets changed (at this time, the only way to change
      // references is by a resize).
      this.variable = variable != null ? variable : getVariable();
      
      // make sure the context work area is initialized
      WorkArea wa = context.get();
      
      // register this for lifetime management
      wa.registerParameter(this);
   }

   /**
    * Get the variable reference, passed to the function/procedure's parameter. Emitted and to be 
    * used only in cases of OUTPUT and INPUT-OUTPUT extent parameters.
    * 
    * @return   See above.
    */
   public abstract T[] getVariable();

   /**
    * Change the caller's reference to the new reference.  Emitted and to be used only in cases of
    * OUTPUT and INPUT-OUTPUT extent parameters.
    * 
    * @param    reference
    *           The new reference.
    */
   public abstract void setVariable(T[] reference);

   /**
    * Initializes the parameter value reference which is used in the function/procedure
    * scope. The method is responsible for its proper initialization and registering
    * the reference in this instance.
    * <p>
    * Use this method to initialize parameter values representing unfixed indeterminate 
    * extents.
    * 
    * @return  A valid and initialized parameter value reference. 
    */
   public abstract T[] initParameter();
   
   /**
    * Initializes the parameter value reference which is used in the function/procedure
    * scope. The method is responsible for its proper initialization and registering
    * the reference in this instance.
    * <p>
    * Use this method to initialize parameter values representing determinate 
    * extents.
    * 
    * @param   extent
    *          The extent size declared by the function/procedure parameter.
    * 
    * @return  A valid and initialized parameter value reference. 
    */
   public abstract T[] initParameter(int extent);
   
   /**
    * Returns the current reference for the parameter.
    *
    * @return  Current parameter reference.    
    */
   public T[] getParameter()
   {
      return parameter;
   }
   
   /**
    * Set the current reference for the parameter. As this is mutable, we need to save it each
    * time the parameter's reference gets changed.
    */
   public void setParameter(T[] parameter)
   {
      this.parameter = parameter;
   }

   /**
    * Assign the current value of the parameter variable back to the caller's variable 
    * reference. This is expected to occur upon exit from the function or procedure 
    * with which this object is associated to.
    */
   private void assign()
   {
      // quick out if an error has occurred
      if (ErrorManager.isPendingError())
      {
         return;
      }
      
      if (parameter == null)
      {
         throw new IllegalStateException("The extent parameter reference was not set," +
         		                     " before the function/procedure end !");
      }
      
      if (parameter != variable)
      {
         // validate the assignment
         boolean valid = validateAssignment();
         if (valid)
         {
            // if the reference has changed and validated ok, 
            // perform the assignment
            performAssignment();
         }
      }
   }
   
   /**
    * An extension point to allow the descendants to provide validation logic
    * for parameter reference to variable assignment on exit from the function 
    * or procedure with which this object is associated to.
    * <p>
    * To indicate a failed validation, indicate an error with help of
    * {@link ErrorManager}.    
    */
   protected boolean validateAssignment()
   {
      // no special validation at this level
      return true;
   }
   
   /**
    * An extension point to allow the descendants to provide custom logic
    * for parameter reference to variable assignment on exit from the function 
    * or procedure with which this object is associated to.
    * <p>
    * The default implementation creates an array copy of the current parameter
    * reference, implements decimal precision management and stores the copy 
    * to the target variable by calling {@link #setVariable(BaseDataType[])}.
    * <p>
    * The method is called only when there is no pending error being indicated 
    * by {@link ErrorManager}.  
    */
   protected void performAssignment()
   {
      // get copy of the latest data
      T[] param = ArrayAssigner.copyOf(parameter);
      
      if (decimal.class.equals(param.getClass().getComponentType()))
      {
         T[] variable = getVariable();
         Integer precision = ArrayAssigner.getDecimalPrecision((decimal[]) variable);
         for (int i = 0; i < param.length; i++)
         {
            // keep the decimal precision of the target variable value.
            // note that the precision on the target variable will become effective
            // only after the next value assignment, which is the expected behavior.
            ((decimal) param[i]).setPrecision(precision);
         }
      }
      
      setVariable(param);
   }
   
   /**
    * Object which manages the context local state of extent parameter wrapper classes.  
    * It manages a deque of scopes, each of which contains zero or more
    * <code>AbstractExtentParameter</code> objects. As the currently executing
    * program's block scopes open and close, these scopes are pushed and
    * popped from the deque.
    */
   private static class WorkArea
   implements Scopeable
   {
      /** Deque containing registered parameter wrappers */
      private Deque<List<AbstractExtentParameter<?>>> scopes = new ArrayDeque<>();
      
      /**
       * Called as each block scope opens, at which time we push a scope onto
       * the deque.
       * 
       * @see com.goldencode.p2j.util.Scopeable#scopeStart()
       */
      public void scopeStart()
      {
         scopes.push(new LinkedList<AbstractExtentParameter<?>>());
      }
      
      /**
       * Called as each block scope closes, at which time we pop a scope from
       * the deque.
       * 
       * @see com.goldencode.p2j.util.Scopeable#scopeFinished()
       */
      public void scopeFinished()
      {
         if (scopes.isEmpty())
         {
            // nothing to do - no params registered for this scope
            return;
         }
         
         // pop the param list for the inner scope
         scopes.pop();
         
         if (scopes.isEmpty())
         {
            // nothing to do - no params registered for this scope
            return;
         }
            
         // param list for the current block
         List<AbstractExtentParameter<?>> eparams = null;
         
         try
         {
            eparams = scopes.peek();
            
            // did the block end with the return error statement?  
            if (BlockManager.getReturnType() == ReturnType.ERROR)
            {
               // prevent the parameter assignment
               return;
            }

            processAssignments(eparams);
         }
         finally
         {
            if (eparams != null)
            {
               eparams.clear();
            }
         }
      }
      
      /**
       * Provides notification that the external procedure scope has been deleted.
       * <p>
       * This is a no-op for output extent parameter implementation.
       */
      @Override
      public void scopeDeleted()
      {
         // no-op
      }

      /**
       * Registers an extent parameter wrapper class to be included in the scope
       * processing.
       * 
       * @param   param
       *          Extent parameter wrapper class.
       */
      public void registerParameter(AbstractExtentParameter<?> param)
      {
         List<AbstractExtentParameter<?>> eparams = null;
         if (scopes.isEmpty())
         {
            // this is the first (and not necessarily the root) scope
            eparams = new LinkedList<>();
            scopes.push(eparams);
         }
         else
         {
            // get the already existing current scope
            eparams = scopes.peek();
         }
         
         eparams.add(param);
      }
      
      /**
       * Cause each (if any) extent parameter wrapper class in the input list 
       * to assign the current value of its parameter value back to the DMO property with
       * which it is associated.
       * 
       * @param   list
       *          The list of extent parameter wrappers.
       */
      private void processAssignments(List<AbstractExtentParameter<?>> list)
      {
         if (list == null || list.isEmpty())
         {
            return;
         }

         boolean batch = (list.size() > 1) && Persistence.isActive();
         
         if (batch)
         {
            RecordBuffer.startBatch();
         }
         
         try
         {
            for (AbstractExtentParameter<? extends BaseDataType> next : list)
            {
               next.assign();
            }
         }
         finally
         {
            if (batch)
            {
               RecordBuffer.endBatch();
            }
         }
      }
   }
}
